
/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */


#include "tbx.h"
#include "marcel.h"
#include "marcel_pmarcel.h"
#include <errno.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/mman.h>


static tbx_htable_t shm_sems;


/* libpthread:
 * -        sem_t: 16 bytes on x86, 32 on ia64,
 * - marcel_sem_t: 16 bytes on x86, 24 on ia64.
 * Good !!!
 */

DEF_MARCEL(void, sem_init, (marcel_sem_t * s, int initial), (s, initial),
{
	/* The code assume that 'initial' is positive.
	 */
	MA_BUG_ON(initial < 0);

	s->value = initial;
	s->first = NULL;
	ma_spin_lock_init(&s->lock);
})
DEF_PMARCEL(int, sem_init, (pmarcel_sem_t *s, int pshared, unsigned int initial), (s,pshared, initial),
{
	MARCEL_LOG_IN();
	if (pshared) {
		MA_NOT_SUPPORTED("pshared semaphore");
		errno = ENOTSUP;
		MARCEL_LOG_RETURN(-1);
	}
	marcel_sem_init(s, initial);
	MARCEL_LOG_RETURN(0);
})
DEF_C(int, sem_init, (pmarcel_sem_t *s, int pshared, unsigned int initial), (s,pshared, initial))

DEF_MARCEL_PMARCEL(int, sem_destroy, (marcel_sem_t *s), (s),
{
	int res = 0;
	ma_spin_lock_bh(&s->lock);
	if (s->first) {
		errno = EBUSY;
		res = -1;
	}
	ma_spin_unlock_bh(&s->lock);
	return res;
})
DEF_C(int, sem_destroy, (pmarcel_sem_t *s), (s))

void marcel_sem_P(marcel_sem_t *s)
{
	semcell c;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	ma_spin_lock_bh(&s->lock);

	if (--(s->value) < 0) {
		c.next = NULL;
		c.blocked = tbx_true;
		c.task = ma_self();
		if (s->first == NULL)
			s->first = s->last = &c;
		else {
			s->last->next = &c;
			s->last = &c;
		}

		INTERRUPTIBLE_SLEEP_ON_CONDITION_RELEASING(c.blocked,
							   ma_spin_unlock_bh_no_resched(&s->lock), ma_spin_lock_bh(&s->lock));
	}
	ma_spin_unlock_bh(&s->lock);

	MARCEL_LOG_OUT();
}

DEF_MARCEL_PMARCEL(int, sem_wait, (marcel_sem_t *s), (s),
{
	semcell c;
	semcell **prev;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	ma_spin_lock_bh(&s->lock);

	if (--(s->value) < 0) {
		c.next = NULL;
		c.blocked = tbx_true;
		c.task = ma_self();
		if (s->first == NULL)
			s->first = s->last = &c;
		else {
			s->last->next = &c;
			s->last = &c;
		}

		MA_SET_INTERRUPTED(0);
		while (c.blocked) {
			ma_set_current_state(MA_TASK_INTERRUPTIBLE);
			ma_spin_unlock_bh_no_resched(&s->lock);
			ma_schedule();
			ma_spin_lock_bh(&s->lock);
			if (MA_GET_INTERRUPTED()
#if defined(MARCEL_DEVIATION_ENABLED) && defined(MA__IFACE_PMARCEL)
			    || c.task->canceled == MARCEL_IS_CANCELED
#endif
				) {
				for (prev = &s->first; *prev && *prev != &c;
				     prev = &(*prev)->next);
				*prev = c.next;
				s->value++;
				ma_spin_unlock_bh(&s->lock);
				errno = EINTR;
				MARCEL_LOG_RETURN(-1);
			}
		}
	}

	ma_spin_unlock_bh(&s->lock);

	MARCEL_LOG_RETURN(0);
})
DEF_C(int, sem_wait, (pmarcel_sem_t *s), (s))

int marcel_sem_try_P(marcel_sem_t *s)
{
	int result = 0;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	ma_spin_lock_bh(&s->lock);
	result = ((s->value - 1) < 0);
	if (! result)
		s->value--;
	ma_spin_unlock_bh(&s->lock);

	MARCEL_LOG_RETURN(!result);
}

DEF_MARCEL_PMARCEL(int, sem_trywait, (marcel_sem_t *s), (s),
{
	if (!marcel_sem_try_P(s)) {
		errno = EAGAIN;
		return -1;
	}
	return 0;
})
DEF_C(int, sem_trywait, (pmarcel_sem_t *s), (s))

/** \brief Attempt to do a timeout bounded P operation on the semaphore
 *
 *  \return 1 if the P operation succeeded before the timeout, 0
 *  otherwise
 */
int marcel_sem_timed_P(marcel_sem_t *s, unsigned long timeout)
{
	unsigned long jiffies_timeout;
	semcell c;
	semcell **prev;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	jiffies_timeout = JIFFIES_FROM_US_ROUNDUP(timeout * 1000);

	ma_spin_lock_bh(&s->lock);
	if (--(s->value) < 0) {
		if (jiffies_timeout == 0) {
			s->value++;
			ma_spin_unlock_bh(&s->lock);
			MARCEL_LOG_OUT();
			return 0;
		}
		c.next = NULL;
		c.blocked = tbx_true;
		c.task = ma_self();
		if (s->first == NULL)
			s->first = s->last = &c;
		else {
			s->last->next = &c;
			s->last = &c;
		}
		do {
			ma_set_current_state(MA_TASK_INTERRUPTIBLE);
			ma_spin_unlock_bh_no_resched(&s->lock);
			jiffies_timeout = ma_schedule_timeout(jiffies_timeout);
			ma_spin_lock_bh(&s->lock);
			if (c.blocked) {
				for (prev = &s->first; *prev && *prev != &c;
				     prev = &(*prev)->next);
				*prev = c.next;
				s->value++;
				ma_spin_unlock_bh(&s->lock);
				MARCEL_LOG_OUT();
				return 0;
			}
		} while (c.blocked);
	} else {
		ma_spin_unlock_bh(&s->lock);
	}

	MARCEL_LOG_OUT();
	return 1;
}

/*******************sem_timedwait**********************/
DEF_MARCEL_PMARCEL(int, sem_timedwait,(marcel_sem_t *__restrict s,
				      const struct timespec *__restrict abs_timeout),(s,abs_timeout),
{
	struct timeval tv;
	long long timeout;	// usec
	unsigned long jiffies_timeout;
	semcell c;
	semcell **prev;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	if (abs_timeout->tv_nsec < 0 || abs_timeout->tv_nsec >= 1000000000) {
		MARCEL_LOG("pmarcel_semtimedwait : valeur nsec(%ld) invalide\n", abs_timeout->tv_nsec);
		errno = EINVAL;
		return -1;
	}

	gettimeofday(&tv, NULL);
	timeout = (abs_timeout->tv_sec - tv.tv_sec) * 1000000 + (abs_timeout->tv_nsec / 1000 - tv.tv_usec);
	if (timeout <= 0)
		jiffies_timeout = 0;
	else
		jiffies_timeout = JIFFIES_FROM_US_ROUNDUP(timeout);

	ma_spin_lock_bh(&s->lock);

	if (--(s->value) < 0) {
		if (jiffies_timeout == 0) {
			s->value++;
			ma_spin_unlock_bh(&s->lock);
			MARCEL_LOG_OUT();
			errno = ETIMEDOUT;
			return -1;
		}
		c.next = NULL;
		c.blocked = tbx_true;
		c.task = ma_self();
		if (s->first == NULL)
			s->first = s->last = &c;
		else {
			s->last->next = &c;
			s->last = &c;
		}
		MA_SET_INTERRUPTED(0);
		do {
			ma_set_current_state(MA_TASK_INTERRUPTIBLE);
			ma_spin_unlock_bh_no_resched(&s->lock);
			jiffies_timeout = ma_schedule_timeout(jiffies_timeout);
			ma_spin_lock_bh(&s->lock);
			if (jiffies_timeout == 0 || MA_GET_INTERRUPTED()) {
				for (prev = &s->first; *prev && *prev != &c;
				     prev = &(*prev)->next);
				*prev = c.next;
				s->value++;
				ma_spin_unlock_bh(&s->lock);
				MARCEL_LOG_OUT();
				errno = MA_GET_INTERRUPTED() ? EINTR : ETIMEDOUT;
				return -1;
			}
		} while (c.blocked);
	} else {
		ma_spin_unlock_bh(&s->lock);
	}

	MARCEL_LOG_RETURN(0);
})
DEF_C(int,sem_timedwait,(sem_t *__restrict sem, const struct timespec *__restrict abs_timeout),(sem,abs_timeout))


void marcel_sem_V(marcel_sem_t *s)
{
	semcell *c;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	ma_spin_lock_bh(&s->lock);

	if (++(s->value) <= 0) {
		c = s->first;
		s->first = c->next;
		ma_spin_unlock_bh(&s->lock);
		c->blocked = tbx_false;
		ma_smp_wmb();
		ma_wake_up_thread(c->task);
	} else {
		ma_spin_unlock_bh(&s->lock);
	}

	MARCEL_LOG_OUT();
}

int marcel_sem_try_V(marcel_sem_t *s)
{
	semcell *c;

	MARCEL_LOG_IN();
	MARCEL_LOG("semaphore %p\n", s);

	if (!(ma_spin_trylock_bh(&s->lock)))
		MARCEL_LOG_RETURN(0);

	if (s->value < 0) {
		c = s->first;
		if (!(ma_wake_up_thread_async(c->task))) {
			ma_spin_unlock_bh(&s->lock);
			MARCEL_LOG_RETURN(0);
		}
		/* This is a bit late, but since we hold the lock, the woken up
		 * thread will spin waiting for us */
		c->blocked = tbx_false;
		s->first = c->next;
	}

	s->value++;
	ma_spin_unlock_bh(&s->lock);

	MARCEL_LOG_RETURN(1);
}

DEF_MARCEL_PMARCEL(int, sem_post, (marcel_sem_t *s), (s),
{
	// TODO: vrifier tout overflow de value
	marcel_sem_V(s);
	return 0;
})
DEF_C(int, sem_post, (pmarcel_sem_t *s), (s))

DEF_MARCEL_PMARCEL(int, sem_getvalue, (marcel_sem_t * __restrict s, int * __restrict sval), (s, sval),
{
	ma_spin_lock_bh(&s->lock);
	*sval = s->value;
	ma_spin_unlock_bh(&s->lock);
	return 0;
})
DEF_C(int, sem_getvalue, (pmarcel_sem_t * __restrict s, int * __restrict sval), (s, sval))


/********************sem_close**************************/
DEF_MARCEL_PMARCEL(int,sem_close,(marcel_sem_t *sem),(sem),
{
	char *semname;
	MARCEL_LOG_IN();

	if (! sem)
		MARCEL_LOG_RETURN(-1);

	/** decrement ref counter **/
	if (0 == (-- *((int *)(sem + 1)))) {
		semname = (char *)sem + sizeof(*sem) + sizeof(int);
		tbx_htable_extract(&shm_sems, semname);
		marcel_munmap(sem, sizeof(*sem) + sizeof(int) + strlen(semname));
	}

	MARCEL_LOG_RETURN(0);
})
DEF_C(int,sem_close,(sem_t *sem),(sem))
DEF___C(int,sem_close,(sem_t *sem),(sem))


/********************sem_open***************************/
DEF_MARCEL(marcel_sem_t*,sem_open,(const char *name, int flags,...),(name,flags,...),
{
	int fd;
	int mode;
	va_list args;
	unsigned int namesz;
	unsigned int value;
	marcel_sem_t *sem;

	MARCEL_LOG_IN();
	
	value = mode = 0;
	if ((flags & O_CREAT)) {
		/** retrieve init semaphore value and access rights **/
		va_start(args, flags);
		mode = va_arg(args, int);
		value = va_arg(args, unsigned int);
		va_end(args);

		if (SEM_VALUE_MAX < value) {
			errno = EINVAL;
			MARCEL_LOG_RETURN(MARCEL_SEM_FAILED);
		}
	}

	/** check semfile permissions & existence **/
	fd = shm_open(name, flags | O_RDWR | O_EXCL, mode | S_IWUSR | S_IRUSR);
	if (-1 == fd && !(flags & O_EXCL)) {
		/* case of unmapped and linked semaphore */
		mode = -1; // mode == -1 + O_CREAT => ignore value, mode, init
		fd = shm_open(name, flags | O_RDWR, 0);
	}
	if (-1 == fd)
		MARCEL_LOG_RETURN(MARCEL_SEM_FAILED);

	/** check if there is a semaphore previously created with this key **/
	sem = tbx_htable_get(&shm_sems, name);
	if (sem) {
		++ *((int *)(sem + 1));
		close(fd);
		MARCEL_LOG_RETURN(sem);
	}
			
	/** map file: | semaphore || ref counter || name_size || name | **/
	namesz = strlen(name);
	if ((flags & O_CREAT) && -1 != mode && 
	    -1 == ftruncate(fd, sizeof(*sem) + sizeof(int) + namesz)) {
		close(fd);
		MARCEL_LOG_RETURN(MARCEL_SEM_FAILED);
	}

	if (NULL == (sem = (marcel_sem_t *)marcel_mmap(NULL, sizeof(*sem) + sizeof(int) + namesz,
						       PROT_READ | PROT_WRITE, MAP_SHARED,
						       fd, 0))) {
		close(fd);
		MARCEL_LOG_RETURN(MARCEL_SEM_FAILED);
	}

	/** initialize and register the semaphore **/
	close(fd);
	tbx_htable_add(&shm_sems, name, sem);
	*((int *)(sem + 1)) = 1;
	if ((flags & O_CREAT) && -1 != mode) {
		marcel_sem_init(sem, value);
		strcpy((char *)sem + sizeof(*sem) + sizeof(int), name);
	}

	MA_WARN_USER("semaphore: does not support process shared lock\n");
	MARCEL_LOG_RETURN(sem);
})
DEF_PMARCEL(marcel_sem_t*,sem_open,(const char *name, int flags,...),(name,flags,...),
{
	if ((flags & O_CREAT)) {
		int perms;
		unsigned int value;
		va_list args;

		va_start(args, flags);
		perms = va_arg(args, int);
		value = va_arg(args, unsigned int);
		va_end(args);

		return marcel_sem_open(name, flags, perms, value);
	}

	return marcel_sem_open(name, flags);
})
DEF_C(sem_t *,sem_open,(const char *name, int flags,...),(name,flags,...))
DEF___C(sem_t *,sem_open,(const char *name, int flags,...),(name,flags,...))

/*******************sem_unlink**************************/
DEF_MARCEL_PMARCEL(int,sem_unlink,(const char *name),(name),
{
	if (0 == shm_unlink(name)) {
		tbx_htable_extract(&shm_sems, name);
		return 0;
	}

	return -1;
})
DEF_C(int,sem_unlink,(const char *name),(name))
DEF___C(int,sem_unlink,(const char *name),(name))



static void sem_data_init(void)
{
	tbx_htable_init(&shm_sems, 10);
}
__ma_initfunc(sem_data_init, MA_INIT_SEM_DATA, "Semaphore static data");
